//
// Copyright 2006-2013 Johannes Hofmann <Johannes.Hofmann@gmx.de>
//
// This software may be used and distributed according to the terms
// of the GNU General Public License, incorporated herein by reference.

#include <pthread.h>
#include "workpool.H"

struct Task {
	action_fn action;
	void *data; 
	Task *prev;
	Task *next;
};

struct Queue {
	Task *head, *tail;
	WorkPool *pool;
};

WorkPool::WorkPool(int numThreads)
{
	queues = new Vector <Queue*> ();
	threads = new Vector <pthread_t> ();
	pthread_mutex_init(&mutex, NULL);
	pthread_mutex_init(&condMutex, NULL);
	pthread_cond_init(&cond, NULL);

	for (int i = 0; i < numThreads; i++) {
		pthread_t t;
		pthread_create(&t, NULL, thread_main, this);
		threads->set(i, t);
	}
}

WorkPool::~WorkPool()
{
	void *res;

	for (int i = 0; i < threads->size (); i++)
		pthread_join(*(threads->get(i)), &res);
	delete queues;
	delete threads;
}

Queue *
WorkPool::createQueue()
{
	Queue *q = (Queue*) malloc(sizeof(Queue));

	q->head = q->tail = NULL;
	q->pool = this;
	
	return q;
}

void
WorkPool::addQueue(Queue *q)
{
	pthread_mutex_lock(&mutex);
	queues->set (queues->size(), q);
	pthread_mutex_unlock(&mutex);
	signal ();
}

void
WorkPool::addTask(Queue *q, action_fn action, void *data)
{
	Task *t = (Task*) malloc (sizeof(Task));

	t->action = action;
	t->data = data;
	t->next = NULL;

	pthread_mutex_lock(&mutex);
	t->prev = q->tail;
	if (!q->head)
		q->head = t;
	if (q->tail)
		q->tail->next = t;
	q->tail = t;
	pthread_mutex_unlock(&mutex);

	signal();
}

void
WorkPool::removeQueues()
{
	pthread_mutex_lock(&mutex);
	queues->truncate (0);
	pthread_mutex_unlock(&mutex);
}

Task*
WorkPool::getTask(Queue *q)
{
	Task *t = q->head;

	if (t) {
		q->head = t->next;
		if (q->head)
			q->head->prev = NULL;
	}

	return t;
}

Task*
WorkPool::getTask()
{
	Task *t = NULL;

	pthread_mutex_lock(&mutex);
	for (int i = 0; i < queues->size(); i++) {
		t = getTask(*queues->get(i));
	 	if (t)
			break;
	}
	pthread_mutex_unlock(&mutex);

	return t;
}

void
WorkPool::signal ()
{
	pthread_mutex_lock(&condMutex);
	pthread_cond_broadcast(&cond);
	pthread_mutex_unlock(&condMutex);
}

void
WorkPool::run ()
{
	Task *t;

	for (;;) {
		pthread_mutex_lock(&condMutex);
		t = getTask ();
		if (!t)
			pthread_cond_wait(&cond, &condMutex);
		pthread_mutex_unlock(&condMutex);

		if (t) {
			t->action(t->data);
			free(t);
		}
	}
}

void *
WorkPool::thread_main(void *data)
{
	WorkPool *pool = (WorkPool*) data;

	pool->run ();
	return NULL;
}
